import pandas as pd
import matplotlib.pyplot as plt
import numpy as np
from matplotlib import gridspec

def extract_stable_data(datafile, measurement_set):
    datapoint = {}
    # df is a pandas data frame
    df = pd.read_csv(datafile)

    # extract phase mean and variance for stable measurements (don't ask what loc means)
    phases = df.loc[(df['EQUILIBRIUM_INDICATOR'] == 31) & (df['SET_NAME'] == measurement_set), 'S21_PHASE']
    if phases.size == 0:
        return None
    datapoint['phase_mean']=phases.mean()
    datapoint['phase_var']=phases.var()

    magnitudes = df.loc[(df['EQUILIBRIUM_INDICATOR'] == 31) & (df['SET_NAME'] == measurement_set), 'S21_MAGNITUDE']
    datapoint['magnitude_mean'] = magnitudes.mean()
    datapoint['magnitude_var'] = magnitudes.var()

    temperatures = df.loc[(df['EQUILIBRIUM_INDICATOR'] == 31) & (df['SET_NAME'] == measurement_set), 'READBACK_TEMPERATURE']
    datapoint['temperature_mean'] = temperatures.mean()
    datapoint['temperature_var'] = temperatures.var()

    humidities = df.loc[(df['EQUILIBRIUM_INDICATOR'] == 31) & (df['SET_NAME'] == measurement_set), 'READBACK_HUMIDITY']
    datapoint['humidity_mean'] = humidities.mean()
    datapoint['humidity_var'] = humidities.var()

    return datapoint


# sweep_type is either 'temperature' or 'humidity'
def plot_sweep(temperatures, humidities, basename, sweep_type, measurement_sets):
    set_data = {}
    for measurement_set in measurement_sets:
        set_data[measurement_set] = {'phases': [], 'phase_vars': [], 'magnitudes': [], 'magnitude_vars': [],
                                     'x_data': []}

    for temp, hum in zip(temperatures, humidities):
        datafile = basename+'_'+str(temp)+'deg_'+str(hum)+'rh.csv'
        print(datafile)
        for measurement_set in measurement_sets:
            datapoint = extract_stable_data(datafile, measurement_set)
            if datapoint is None:
                continue

            data = set_data[measurement_set]
            if sweep_type == 'temperature':
                data['x_data'].append(datapoint['temperature_mean'])
            elif sweep_type == 'humidity':
                data['x_data'].append(datapoint['humidity_mean'])
            else:
                raise Exception('Unknown sweep_type:'+str(sweep_type))

            data['phases'].append(datapoint['phase_mean'])
            data['phase_vars'].append(datapoint['phase_var'])
            data['magnitudes'].append(datapoint['magnitude_mean'])
            data['magnitude_vars'].append(datapoint['magnitude_var'])

    #fig = plt.figure()
    #gs = fig.add_gridspec(2, 1, hspace=0)
    #(ax1, ax2) = gs.subplots(sharex=True)
    fig, (ax1, ax2) = plt.subplots(2, sharex=True, gridspec_kw={'hspace': 0}, figsize=(10, 5))
    ax1.tick_params(bottom=True, top=True, direction='in')
    ax2.tick_params(bottom=True, top=True, direction='inout')

    if sweep_type == 'temperature':
        fig.suptitle(basename + ': Temperature sweep at ' + str(humidities[0]) + ' % r.h.')
        ax2.set_xlabel('temperature [deg C]')
    elif sweep_type == 'humidity':
        fig.suptitle(basename + ': Humidity sweep at ' + str(temperatures[0]) + ' deg C')
        ax2.set_xlabel('relative humidity [%]')
    else:
        raise Exception('Unknown sweep_type:'+str(sweep_type))

    for measurement_set in measurement_sets:
        ax1.errorbar(set_data[measurement_set]['x_data'],
                     set_data[measurement_set]['phases'], set_data[measurement_set]['phase_vars'],
                     label=measurement_set, marker='+', linewidth=0)
    ax1.set_ylabel('S21 phase [deg]')

    for measurement_set in measurement_sets:
        ax2.errorbar(set_data[measurement_set]['x_data'],
                     set_data[measurement_set]['magnitudes'], set_data[measurement_set]['magnitude_vars'],
                     label=measurement_set, marker='+', linewidth=0)
    ax2.set_ylabel('S21 magnitude [dB]')

    # Remove the highest tick label from the lower subplot. It probably overlaps with the lowest one of the upper plot.
    yticks = ax2.yaxis.get_major_ticks()
    yticks[-1].set_visible(False)
    yticks[-2].label1.set_visible(False)  # don't ask me why we have to set the last two tick marks to invisible
    ax2.legend(loc='center right', bbox_to_anchor=(1.5, 1.0))

    fig.tight_layout()  # otherwise the legend is clipped
    plt.subplots_adjust(top=0.9)  # avoid overlap with suptitle

    fig.savefig(basename+measurement_set+'_analysis.pdf')
    
    plt.show()


if __name__ == '__main__':
    print('run \'./prototype.py -t first_tempsweep.txt -p -o tempsweep1\' to get the data needed for this plot.' )
    plot_sweep(np.arange(25., 31.+1.), [35.]*7, 'tempsweep1', 'temperature',
               ['1.3GHz', '1.0GHz', '3.0GHz', '6.0GHz', '10.0GHz'])
