#!/usr/bin/python3
import csv
import time
import numpy
from argparse import ArgumentParser
import pandas as pd
import matplotlib.pyplot as plt
import climate_chamber
import vna_measurement
import virtual_time
import json
import MeasurementPlot
import sys
import analysis
import external_sensors
import PostPlot
import os

# Only use these when calculating the equilibrium indicator. Don't use in the algorithm.
TEMPERATURE_STABLE = 0x1
HUMIDITY_STABLE = 0x2
DUT_SIGNAL0_STABLE = 0x4
DUT_SIGNAL1_STABLE = 0x8
MEASUREMENT_STABLE = 0x10


# Class has only attributes which we are using in the read_data_function to read data from DUT and Chamber
class MeasurementData:
    
    def __init__(self, timestamp, temp, hum, dut_data, percent_temp_heater,
                 percent_hum_heater, temp_dut, temp_room, temp_meas_instr, hum_dut, hum_room,
                 hum_meas_instr, air_press_room, temp_chamber_meas_instr, hum_chamber_meas_instr):
        self.timestamp = timestamp
        self.temp = temp
        self.hum = hum
        self.dut_data = dut_data

        self.percent_temp_heater = percent_temp_heater
        self.percent_hum_heater = percent_hum_heater
        self.temp_dut = temp_dut
        self.temp_room = temp_room
        self.temp_meas_instr = temp_meas_instr
        self.temp_chamber_meas_instr=temp_chamber_meas_instr
        self.hum_dut = hum_dut
        self.hum_room = hum_room
        self.air_press_room = air_press_room
        self.hum_meas_instr = hum_meas_instr
        self.hum_chamber_meas_instr = hum_chamber_meas_instr


class Measurements:
    def __init__(self, config_data, output_basename, standby, ext_sensor_channels):
        self.max_delta_temp = config_data['delta_temp']
        self.max_delta_hum = config_data['delta_hum']

        self.ext_sensor_channels = ext_sensor_channels
        self.sleep_time = config_data["sleep_time"]
        target_accuracy = [self.max_delta_temp, self.max_delta_hum]
        self.chamber = climate_chamber.create_chamber(config_data['chamber_ip'], target_accuracy)
        self.instr_chamber = climate_chamber.create_chamber(config_data['instr_chamber_ip'],
                                                            target_accuracy)
        # FIXME: Do we want a factory function for the DUTs?
        if config_data['dut']['type'] == 'VNA':
            self.dut = vna_measurement.VnaMeasurement(config_data['dut'], target_accuracy)
        else:
            raise Exception('Unknown DUT type: '+config_data['dut']['type'])

        self.max_delta_dut_signals = self.dut.get_dut_max_delta_signals()

        # data logger for external sensors
        self.ext_sensors = external_sensors.create_sensors(config_data['logger_model'], config_data['logger_ip'],
                                                           ext_sensor_channels)

        self.standby = standby
        self.output_basename = output_basename
        self.clock = virtual_time.get_clock(config_data['chamber_ip'], target_accuracy)
        self.temperature_stable = False
        self.humidity_stable = False
        self.dut_signals_stable = [False, False]

        self.postplot_obj = None
        
        self.measurement_plot = MeasurementPlot.MeasurementPlot(trace_subplot5=config_data['trace_subplot5'])
        self.data_collection_for_online_plot = []

    def perform_measurements(self, sweep_file):
        with open(sweep_file) as file:
            try:
                measurement_number = 0
                while line := file.readline().rstrip():
                    list_of_values = line.split()
                    next_temp = float(list_of_values[0])  # target_temperature
                    next_hum = float(list_of_values[1])  # target_humidity
                    next_soaking = int(list_of_values[2])  # soaking_time
                    next_reads = int(list_of_values[3])  # number of stable readings

                    self.measurement_plot.fig.suptitle("Measurement "+str(measurement_number)+': ' + str(next_temp) +
                                                       ' degC, ' + str(next_hum) + ' rel. hum.', color="red")
                    self.perform_single_measurement(self.output_basename+'_'+str(measurement_number)+'.csv', next_temp,
                                                    next_hum, next_soaking, next_reads)
                    measurement_number += 1

            except KeyboardInterrupt:
                pass
            except MeasurementPlot.PlottingError:
                # Just the plotting failed, probably because the window was closed
                # The measurement was partly done and should be plotted, so the following code has to see the correct
                # number of successful measurements
                measurement_number += 1

        plt.close()

        if self.standby:
            standby_response = self.chamber.set_mode('STANDBY')
            print(standby_response)

        return measurement_number

    def perform_sweep(self, start_val, stop_val, step_val, const_param, soaking_time, n_stable_reads, sweep_type):
        current_val = start_val
        step_sign = numpy.sign(stop_val-start_val)
        sweep_values = []
        while step_sign*current_val < step_sign*stop_val:
            sweep_values.append(current_val)
            current_val = round(current_val + step_sign*step_val, 1)
        sweep_values.append(stop_val)

        successful_sweep_values = []
        try:
            for val in sweep_values:
                if sweep_type == 'temperature':
                    target_temp = val
                    target_hum = const_param
                elif sweep_type == 'humidity':
                    target_temp = const_param
                    target_hum = val
                else:
                    raise Exception('Unknown sweep type: '+sweep_type)

                output_name = self.output_basename + '_'+str(target_temp)+'deg_'+str(target_hum)+'rh.csv'
                self.measurement_plot.fig.suptitle(sweep_type + " sweep measurement "+str(sweep_values.index(val)+1)+
                                                   '/'+str(len(sweep_values))+': ' + str(target_temp)+' degC, '
                                                   + str(target_hum) + ' rel. hum.',
                                                   color="red")
                self.perform_single_measurement(output_name, target_temp, target_hum, soaking_time, n_stable_reads)

        except KeyboardInterrupt:
            pass
        except MeasurementPlot.PlottingError:
            # Remove the remaining measurements from the list.
            # One measurement was partly done and should be plotted, so we leave 'val' in the list
            del sweep_values[sweep_values.index(val)+1:]

        plt.close()
        return sweep_values

    def perform_single_measurement(self, output, target_temp, target_hum, soaking_time, n_stable_reads):
        """
        A "single measurement refers to a single temperature and humidity setting.
        This is a "chamber measurement point". It consists out of multiple measurements sets. The data for all
        measurement sets of this chamber point are taken for this measurement.
        """
        with open(output, mode='w', newline='') as csv_file:
            fieldnames = ['TIMESTAMP', 'TARGET_TEMPERATURE', 'READBACK_TEMPERATURE', 'TARGET_HUMIDITY',
                          'READBACK_HUMIDITY', 'DUT_IDENTIFIER', 'RUN_ID', 'EQUILIBRIUM_INDICATOR', 'TEMP_HEATER',
                          'HUM_HEATER', 'TEMP_DUT', 'TEMP_ROOM', 'TEMP_MEAS_INSTR', 'HUM_DUT', 'HUM_ROOM',
                          'HUM_MEAS_INSTR', 'AIR_PRESS_ROOM', 'READBACK_TEMP_MEAS_INSTR', 'READBACK_HUM_MEAS_INSTR']
            fieldnames.extend(self.dut.get_dut_signal_names())

            # csv.dict writer add adda row wise
            writer = csv.DictWriter(csv_file, fieldnames=fieldnames)
            writer.writeheader()
            self.data_collection_for_online_plot = []
            plt.ion()

            set_const_response = self.chamber.set_const((target_temp, target_hum))
            print(set_const_response)
            set_mode_response = self.chamber.set_mode('CONSTANT')
            print(set_mode_response)

            number_of_soaking_reads = soaking_time / self.sleep_time + 1

            do_another_measurement = True
            # next_read_time is the starttime of the second read (i.e. read after this one). The time of
            # this read (i.e. the first read in the measurement) is now().
            next_read_time = self.clock.time() + self.sleep_time
            while do_another_measurement:
                # wait until set point is reached (+soaking time)
                dut_signal_queues = [[], []]

                while True:
                    # Only read the climate chamber data once. The equilibrium indicator is only calculated on the
                    # primary data set.
                    data = self.read_data(self.dut.get_measurement_set_names()[0])
                    # if it is not within the target range reset start time
                    self.temperature_stable = False
                    self.humidity_stable = False

                    self.temperature_stable = self.calculate_temperature_stability(target_temp, float(data.temp))
                    self.humidity_stable = self.calculate_humidity_stability(target_hum, float(data.hum))

                    # Use indexed loop because we need to modify, and zip seems to copy the content :-(
                    for i, signal_queue in enumerate(dut_signal_queues):
                        # The queue must not be longer than the max number of soaking reads.
                        # If the queue is already full, we have to pop the first element before we can add the
                        # current measurement.
                        if len(signal_queue) >= number_of_soaking_reads:
                            signal_queue.pop(0)
                        if self.temperature_stable and self.humidity_stable:
                            signal_queue.append(data.dut_data[self.dut.get_dut_reference_signal_names()[i]])
                        else:
                            signal_queue.clear()

                        self.dut_signals_stable[i] = False
                        if len(signal_queue) >= number_of_soaking_reads:
                            spread = max(signal_queue) - min(signal_queue)
                            if spread < 2*self.max_delta_dut_signals[i]:
                                self.dut_signals_stable[i] = True

                    print('Setpoint: ' + str(target_temp) + ' ' + str(target_hum) + ' | Temp: ' + data.temp +
                          ' °C' + ' | Humid: ' + data.hum + '%'
                          + ' | soaking read nr' + str(len(dut_signal_queues[0])))
                    self.store_and_plot_data(target_temp, target_hum, data, self.cook_up_equi_indicator())
                    writer.writerow(self.data_collection_for_online_plot[-1])

                    # read and store data for all further sets
                    for set_name in self.dut.get_measurement_set_names()[1:]:
                        data.dut_data = self.dut.get_dut_measurements(set_name)
                        writer.writerow(self.create_data_dictionary(target_temp, target_hum, data,
                                                                    self.cook_up_equi_indicator()))

                    if self.temperature_stable and self.humidity_stable and self.dut_signals_stable[0] and\
                            self.dut_signals_stable[1]:
                        reference_values = []
                        for signal_queue in dut_signal_queues:
                            reference_values.append(signal_queue[-1])
                        print('SOAKING FINISHED!')
                        break
                    else:
                        self.sleep_until(next_read_time)
                        next_read_time += self.sleep_time

                # perform the configured number of measurements and check that they are really stable
                # It started running after everything become stable
                supposedly_stable_measurements = {}
                for set_name in self.dut.get_measurement_set_names():
                    supposedly_stable_measurements[set_name] = []
                all_measurements_stable = True

                for i in range(0, n_stable_reads):
                    # create the main data set for the first measurement set
                    data = self.read_data(self.dut.get_measurement_set_names()[0])
                    self.temperature_stable = self.calculate_temperature_stability(target_temp, float(data.temp))
                    self.humidity_stable = self.calculate_humidity_stability(target_hum, float(data.hum))
                    this_measurement_stable = (self.temperature_stable and self.humidity_stable)
                    for j, max_delta in enumerate(self.max_delta_dut_signals):
                        value = data.dut_data[self.dut.get_dut_reference_signal_names()[j]]
                        self.dut_signals_stable[j] = (reference_values[j] - max_delta <= value) and\
                            (value <= reference_values[j] + max_delta)
                        if not self.dut_signals_stable[j]:
                            this_measurement_stable = False

                    self.store_and_plot_data(target_temp, target_hum, data, self.cook_up_equi_indicator())
                    supposedly_stable_measurements[self.dut.get_measurement_set_names()[0]].append(
                        self.data_collection_for_online_plot[-1])

                    # read and store data for all further sets
                    for set_name in self.dut.get_measurement_set_names()[1:]:
                        data.dut_data = self.dut.get_dut_measurements(set_name)
                        supposedly_stable_measurements[set_name].append(
                            self.create_data_dictionary(target_temp, target_hum, data, self.cook_up_equi_indicator()))
                                
                    if this_measurement_stable:
                        print('Stable measurement ' + str(i+1) + '/' + str(n_stable_reads))
                        self.sleep_until(next_read_time)
                        next_read_time += self.sleep_time
                    else:
                        all_measurements_stable = False
                        print('Measurement not stable. Retrying.')
                        break

                for set_name in self.dut.get_measurement_set_names():
                    for measurement in supposedly_stable_measurements[set_name]:
                        if all_measurements_stable:
                            measurement['EQUILIBRIUM_INDICATOR'] = TEMPERATURE_STABLE | HUMIDITY_STABLE | \
                                                                   DUT_SIGNAL0_STABLE | DUT_SIGNAL1_STABLE | \
                                                                   MEASUREMENT_STABLE
                            do_another_measurement = False

                        writer.writerow(measurement)

    def sleep_until(self, wakeup_time):
        remaining_sleep_time = wakeup_time - self.clock.time()
        if remaining_sleep_time > 0:
            self.clock.sleep(remaining_sleep_time)

    def read_data(self, measurement_set_name):
        [temp, hum, mode, alarms] = self.chamber.read_monitor().split(',')
        perc_temp_heater, perc_hum_heater = self.chamber.get_heater_percentage()

        dut_data = self.dut.get_dut_measurements(measurement_set_name)

        temp_dut, temp_room, temp_meas_instr, hum_dut, hum_room, hum_meas_instr, air_press_room = \
            self.ext_sensors.get_sensor_values()
        
        [temp_chamber_meas_instr, hum_chamber_meas_instr, mode_meas_instr, alarms_meas_instr] = \
            self.instr_chamber.read_monitor().split(',')

        return MeasurementData(int(self.clock.time()), temp, hum, dut_data,
                               perc_temp_heater, perc_hum_heater, temp_dut, temp_room, temp_meas_instr,
                               hum_dut, hum_room, hum_meas_instr, air_press_room, temp_chamber_meas_instr,
                               hum_chamber_meas_instr)

    def create_data_dictionary(self, target_temp, target_hum, data, equi_indicator):
        """
        Create a dictionary out of the  MeasurementData object and some extra data, as needed by the writer
        """
        measurement = {
            'TIMESTAMP': data.timestamp,
            'TARGET_TEMPERATURE': target_temp,
            'READBACK_TEMPERATURE': float(data.temp),
            'TARGET_HUMIDITY': target_hum,
            'READBACK_HUMIDITY': float(data.hum),
            'EQUILIBRIUM_INDICATOR': equi_indicator,
            'TEMP_HEATER': data.percent_temp_heater,
            'HUM_HEATER': data.percent_hum_heater,
            'TEMP_DUT': data.temp_dut,
            'TEMP_ROOM': data.temp_room,
            'TEMP_MEAS_INSTR': data.temp_meas_instr,
            'HUM_DUT': data.hum_dut,
            'HUM_ROOM': data.hum_room,
            'HUM_MEAS_INSTR': data.hum_meas_instr,
            'AIR_PRESS_ROOM': data.air_press_room,
            'READBACK_TEMP_MEAS_INSTR': float(data.temp_chamber_meas_instr),
            'READBACK_HUM_MEAS_INSTR': float(data.hum_chamber_meas_instr),
        }
        measurement.update(data.dut_data)
        return measurement

    def store_and_plot_data(self, target_temp, target_hum, data, equi_indicator):
        measurement = self.create_data_dictionary(target_temp, target_hum, data, equi_indicator)
        self.data_collection_for_online_plot.append(measurement)
        data_frame = pd.DataFrame(self.data_collection_for_online_plot)
        self.measurement_plot.draw(data_frame)

    def current_milli_time(self):
        return int(round(self.clock.time() * 1000))

    def cook_up_equi_indicator(self):
        equilibrium_indicator = 0

        if self.temperature_stable:
            equilibrium_indicator = equilibrium_indicator | TEMPERATURE_STABLE

        if self.humidity_stable:
            equilibrium_indicator = equilibrium_indicator | HUMIDITY_STABLE

        if self.dut_signals_stable[0]:
            equilibrium_indicator = equilibrium_indicator | DUT_SIGNAL0_STABLE

        if self.dut_signals_stable[1]:
            equilibrium_indicator = equilibrium_indicator | DUT_SIGNAL1_STABLE

        return equilibrium_indicator

    def calculate_temperature_stability(self, target_temp, readback_temp):
        return (target_temp-self.max_delta_temp <= float(readback_temp)) and \
               (float(readback_temp) <= target_temp+self.max_delta_temp)

    def calculate_humidity_stability(self, target_hum, readback_hum):
        return (target_hum-self.max_delta_hum <= float(readback_hum)) and \
               (float(readback_hum) <= target_hum+self.max_delta_hum)


def plot_output(output_basename, measurements_appendices, show_blocking_plot, config_data, ext_sens_data,
                measurement_set_names, title=''):
    
    time_unit = str(config_data['time_unit'])

    list_of_frames = []
    storepath = os.path.join(os.getcwd(), 'PostPlots')

    post_plot = PostPlot.PostPlot(trace_subplot5=config_data['trace_subplot5'])
    for index, m in enumerate(measurements_appendices):
        measurement_name = output_basename+'_'+str(m)
        list_of_frames.append(pd.read_csv(measurement_name+'.csv'))

        for set_name in measurement_set_names:
            post_plot.plot_frame_data(list_of_frames[-1], title+set_name, time_unit, set_name)
            post_plot.save_fig(storepath, measurement_name+set_name+'.pdf')

    combined_data_frame = pd.concat(list_of_frames, ignore_index=True, sort=False)
    
    for set_name in measurement_set_names:
        post_plot.plot_frame_data(combined_data_frame, title+set_name, time_unit, set_name)
        post_plot.save_fig(storepath, output_basename+set_name + '.pdf')
    
    if show_blocking_plot:
        plt.ioff()
    
    plot = MeasurementPlot.MeasurementPlot(title = title, trace_subplot5 = \
                                           config_data['trace_subplot5'] )
    plot.draw(combined_data_frame, output_basename + '_graph.pdf')


def run_temperature_sweep_from_file(temperature_sweep_file, meas):
    with open(temperature_sweep_file) as file:
        try:
            line = file.readline().rstrip()
            list_of_values = line.split()
            start_temp = float(list_of_values[0])
            stop_temp = float(list_of_values[1])
            temperature_step = float(list_of_values[2])
            target_hum = float(list_of_values[3])
            soaking_time = int(list_of_values[4])
            n_stable_reads = int(list_of_values[5])

            return meas.perform_sweep(start_temp, stop_temp, temperature_step, target_hum, soaking_time,
                                      n_stable_reads, 'temperature'), target_hum

        except KeyboardInterrupt:
            pass

if __name__ == '__main__':
    parser = ArgumentParser()
    parser.add_argument('-f', '--file',
                        help='File containing custom list of measurements',
                        default='')
    parser.add_argument('-t', '--temperaturesweepfile',
                        help='File containing temperature sweep parameters',
                        default='')
    parser.add_argument('-o', '--output',
                        help='Name of output file')
    parser.add_argument("-s", "--standby",
                        action="store_true", default=False,
                        help="Change Chamber mode to standby at the end")
    parser.add_argument("-p", "--plot",
                        action="store_true", default=False,
                        help="Show blocking plot window at the end")

    # parser.add_argument("-q", "--quiet",
    #                     action="store_false", dest="verbose", default=True,
    #                     help="don't print status messages to stdout")
    args = parser.parse_args()

    if args.file == '' and args.temperaturesweepfile == '':
        print('Argument error: Either \'file\' or \'temperaturesweepfile\' must be specified.')
        sys.exit(-1)

    if not args.output:
        output_basename = time.strftime("%Y_%m_%d-%H_%M_%S") + "_results"
    else:
        output_basename = args.output

    print(args.file, output_basename, args.standby)

    # reading json file for target accuracy
    with open('test_stand_parameter.json', 'r') as f:
        config_data = json.load(f)

    with open('ext_sensor_channels.json', 'r') as f2:
        ext_sensor_channels = json.load(f2)

    mes = Measurements(config_data, output_basename, args.standby, ext_sensor_channels)
    try:
        if args.file:
            n_measurements = mes.perform_measurements(args.file)
            plot_output(output_basename, range(n_measurements), args.plot, config_data, ext_sensor_channels,
                        mes.dut.get_measurement_set_names(), output_basename)
        if args.temperaturesweepfile:
            temperatures, humidity = run_temperature_sweep_from_file(args.temperaturesweepfile, mes)
            # run analysis here
            temp_extensions = []
            for t in temperatures:
                temp_extensions.append(str(t)+'deg_'+str(humidity)+'rh')
            analysis.plot_sweep(temperatures, [humidity]*len(temperatures), output_basename, 'temperature',
                                mes.dut.get_measurement_set_names())
            plot_output(output_basename, temp_extensions, args.plot, config_data, ext_sensor_channels,
                        mes.dut.get_measurement_set_names(),
                        output_basename + ': Temperature sweep ' +
                        str(temperatures[0]) + ' degC to ' + str(temperatures[-1]) + ' degC')
    finally:
        mes.chamber.close()
        mes.ext_sensors.close()

    # mes.plot_output(output)
